/************************************/
/*----------- GSTEVN.C -------------*/
/* Module de gestion des vnements */
/************************************/
#include  <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>

#include "..\tools\stdprog.h"
#include "..\tools\stdinput.h"
#include "..\tools\xgem.h"
#include "..\tools\gwindows.h"
#include "..\tools\rasterop.h"
#include "..\tools\rzoom.h"
#include "..\tools\chkstack.h"
#include "..\tools\logging.h"

#include     "defs.h"
#include     "undo.h"
#include     "snap.h"
#include   "walbum.h"
#include   "rtzoom.h"
#include  "touches.h"
#include  "actions.h"
#include  "fullscr.h"
#include  "figures.h"
#include "visionio.h"
#include "filtrage.h"
#include "vtoolbar.h"
#include "hexdump.h"

#include "forms\fgeneral.h"
#include "forms\ftexte.h"
#include "forms\fformes.h"
#include "forms\fgomme.h"
#include "forms\fpref.h"
#include "forms\fbaratin.h"
#include "forms\fpencil.h"
#include "forms\faerosol.h"
#include "forms\fbloc.h"
#include "forms\falbum.h"
#include "forms\fsym.h"
#include "forms\fhisto.h"
#include "forms\ftaille.h"
#include "forms\fpal.h"

static int   last_icon ;
static int   zooms[]  = { 1, 2, 4, 8, 16 } ;
static float zratio[] = { 1.0, 2.0, 4.0, 8.0, 16.0 } ;

static int pm(int n, long max)
{
  return (int) ((1000L*(long)n)/max) ;
}

int pmfr(int x, int y)
{
  return (int) (0.5+1000.0*(float)x/(float)y) ;
}

int pmf(int x, int y)
{
  return (int) (1000.0*(float)x/(float)y) ;
}

int pml(int x, int w)
{
  return (int) (((long)x*(long)w)/1000L) ;
}

void update_view(GEM_WINDOW* gw)
{
  VXIMAGE* vimage ;
  int      x, y, w, h ;
  int      rc, rl ;
  int      xi, yi, wx, wy ;
  int      index ;

  if ( gw == NULL ) return ;
  vimage = (VXIMAGE*) gw->Extension ;

  for ( index = 0; index < 4; index++ )
    if ( vimage->zoom_level == zooms[index] ) break ;
  gw->GetWorkXYWH( gw, &x, &y, &w, &h ) ;
  wx = w/zooms[index] ;
  wy = h/zooms[index] ;
  GWGetHSlider( gw, &xi ) ;
  GWGetVSlider( gw, &yi ) ;
  rc = pml( xi, vimage->inf_img.mfdb.fd_w-wx ) ;
  rl = pml( yi, vimage->inf_img.mfdb.fd_h-wy ) ;
  vimage->x1 = rc ;      vimage->y1 = rl ;
  vimage->x2 = rc+wx-1 ; vimage->y2 = rl+wy-1 ;
}

void update_zview(GEM_WINDOW* gw)
{
  VXIMAGE* vimage ;
  OBJECT*  obj ;
  float    echx, echy, ech ;
  char     flag[1+NB_TOOLS] ;

  memzero(flag, 1+NB_TOOLS) ;
  flag[ZOOM_BOX] = 1 ;

  if ( gw == NULL )
  {
    VToolBarUpdate( flag ) ;
    return ;
  }

  vimage = (VXIMAGE*) gw->Extension ;

  echx = (float)adr_icones[ZOOM_BOX].ob_width/(float)vimage->inf_img.mfdb.fd_w ;
  echy = (float)adr_icones[ZOOM_BOX].ob_height/(float)vimage->inf_img.mfdb.fd_h ;
  if (echx < echy) ech = echx ;
  else             ech = echy ;
  vimage->zw  = (int) ((float)vimage->inf_img.mfdb.fd_w*ech) ;
  vimage->zh  = (int) ((float)vimage->inf_img.mfdb.fd_h*ech) ;
  vimage->zx  = (adr_icones[ZOOM_BOX].ob_width-vimage->zw)>>1 ;
  vimage->zy  = (adr_icones[ZOOM_BOX].ob_height-vimage->zh)>>1 ;
  vimage->zzw = (int) ((float)(vimage->x2-vimage->x1)*ech) ;
  vimage->zzh = (int) ((float)(vimage->y2-vimage->y1)*ech) ;
  vimage->zzx = (int) ((float)vimage->x1*ech) ;
  vimage->zzy = (int) ((float)vimage->y1*ech) ;

  obj = &adr_icones[ZOOM_IN] ;
  obj->ob_x      = vimage->zx ;
  obj->ob_y      = vimage->zy ;
  obj->ob_width  = vimage->zw ;
  obj->ob_height = vimage->zh ;

  obj = &adr_icones[ZOOM_VUE] ;
  obj->ob_x      = vimage->zzx ;
  obj->ob_y      = vimage->zzy ;
  obj->ob_width  = vimage->zzw ;
  obj->ob_height = vimage->zzh ;
  VToolBarUpdate( flag ) ;
}

void decheck_moutils(void)
{
  char items[] = { M_OPOINT, M_OTRACE, M_OLIGNE, M_ORECTANGLE, M_OGOMME, M_OPLAN, M_OCERCLE, M_ODISQUE, M_OPEINTURE,
                   M_OTEXTE, M_OPINCEAU, M_OAEROSOL, M_OMAIN, M_OPALETTE, M_OKLIGNE, M_OTABLEAU, M_OBLOC, M_OLOUPE, M_OTEMP } ;
  short i ;

  for ( i = 0; i < ARRAY_SIZE(items); i++ )
    GWMenuUncheck( items[i] ) ;
}

void set_goodmouse(void)
{
  if ( !config.souris_viseur )
  {
    if ( mouse.gr_monumber != ARROW )
    {
      mouse.gr_monumber = ARROW ;
      set_mform(&mouse) ;
    }
  }
  else
  {
    if ( mouse.gr_monumber != USER_DEF )
    {
      mouse.gr_monumber = USER_DEF ;
      CopyMForm( &mouse.gr_mof, &mf_viseur.gr_mof ) ;
      set_mform(&mouse) ;
    }
    else if ( memcmp( &mouse.gr_mof, &mf_viseur.gr_mof, sizeof(MFORM) ) )
    {
      mouse.gr_monumber = USER_DEF ;
      CopyMForm( &mouse.gr_mof, &mf_viseur.gr_mof ) ;
      set_mform(&mouse) ;
    }
  }
}

int ztool(int num)
{
  switch(num)
  {
    case PALETTE     :
    case TEMPERATURE :
    case POINT       :
    case LOUPE       :
    case ZOOM_VUE    : return 1 ;
    case PEINTURE    : if ( ftype.sindex == 8 ) return 1 ;
                       /* Le remplissage est autoris en zoom s'il */
                       /* Le motif  afficher est plein            */
  }

  return 0 ;
}

void traite_move(GEM_WINDOW* gwnd)
{
  VXIMAGE* vimage ;
  float    echx, echy ;
  float    x, y ;
  int      w, h ;
  int      ix, iy, fx, fy ;
  int      lx, ly, lw, lh ;
  int      xi, yi, wx, wy ;
  int      index ;
  int      new ;

  if ( gwnd == NULL ) return ;
  vimage = (VXIMAGE*) gwnd->Extension ;

  for (index = 0; index < 4; index++)
    if ( vimage->zoom_level == zooms[index] ) break ;

  objc_offset(adr_icones, ZOOM_IN, &lx, &ly) ;
  objc_offset(adr_icones, ZOOM_VUE, &ix, &iy) ;
  w  = adr_icones[ZOOM_VUE].ob_width ;
  h  = adr_icones[ZOOM_VUE].ob_height ;
  lw = adr_icones[ZOOM_IN].ob_width ;
  lh = adr_icones[ZOOM_IN].ob_height ;
  my_dragbox(w, h, ix, iy, lx, ly, lw, lh, &fx, &fy) ;
  echx = (float)(vimage->inf_img.mfdb.fd_w) ;
  echx = echx/(float)lw ;
  echy = (float)(vimage->inf_img.mfdb.fd_h) ;
  echy = echy/(float)lh ;
  gwnd->GetWorkXYWH( gwnd, &xi, &yi, &wx, &wy ) ;
  wx = wx/zooms[index] ;
  wy = wy/zooms[index] ;
  x  = (float)(fx-lx)*echx ;
  vimage->x1 = (int) (0.5+x) ;
  if (vimage->x1 < 0) vimage->x1 = 0 ;
  vimage->x2 = vimage->x1+wx-1 ;
  y = (float)(fy-ly)*echy ;
  vimage->y1 = (int)(0.5+y) ;
  if (vimage->y1 < 0) vimage->y1 = 0 ;
  vimage->y2 = vimage->y1+wy-1 ;
  if (vimage->y2 > vimage->inf_img.mfdb.fd_h-1)
    vimage->y2 = vimage->inf_img.mfdb.fd_h-1 ;
  adr_icones[ZOOM_VUE].ob_x = fx-lx ;
  adr_icones[ZOOM_VUE].ob_y = fy-ly ;
  new = pmfr( vimage->x1, vimage->inf_img.mfdb.fd_w-wx ) ;
  GWSetHSlider( gwnd, new ) ;
  new = pmfr( vimage->y1, vimage->inf_img.mfdb.fd_h-wy ) ;
  GWSetVSlider( gwnd, new ) ;
  if ( vimage->zoom_level != 1 )
    traite_loupe( gwnd, vimage->zoom_level, -1, -1 ) ;
  GWRePaint( gwnd ) ;
}

void select_icon(int mx, int my)
{
  VXIMAGE*    vimage ;
  GEM_WINDOW* gwnd = GemApp.CurrentGemWindow ;
  MOUSE_DEF   save ;
  int         num, is_vimg ;

  if ( gwnd == NULL ) return ;
  num = objc_find( adr_icones, 0, 200, mx, my ) ;
  if ( (num == ZOOM_IN) || (num == ZOOM_BOX) ) return ;
  is_vimg = ( gwnd && (strcmp( gwnd->ClassName, VIMG_CLASSNAME ) == 0) ) ;
  vimage  = (VXIMAGE*) gwnd->Extension ;
  if ( is_vimg && (vimage->zoom_level != 1) )
  {
    if ( !ztool( num ) )
    {
      form_stop(  MsgTxtGetMsg(hMsg, MSG_NOTOOLZ)  ) ;
      return ; /* Outil non autoris en zoom */
    }
  }

  if ( num != -1 )
  {
    int  i ;
    char flag[1+NB_TOOLS] ;
    
    for ( i = 1; i <= NB_TOOLS; i++ )
      if ( selected( adr_icones, i ) ) break ;

    if ( i == 1+NB_TOOLS ) i = num ;
    deselect( adr_icones, i ) ;
    if ( i != TEMPERATURE ) last_icon = i ;
    if ( last_icon == TEMPERATURE ) last_icon = POINT ;
    select( adr_icones, num ) ;
    memzero( flag, 1+NB_TOOLS ) ;
    flag[i] = flag[num] = 1 ;
    VToolBarUpdate( flag ) ;
    if ( (num != TEMPERATURE) && (num != PALETTE) ) decheck_moutils() ;
    switch( num )
    {
      case GOMME       : mouse.gr_monumber = USER_DEF ;
                         if (!config.souris_viseur) CopyMForm( &mouse.gr_mof, &mf_gomme.gr_mof ) ;
                         else                       CopyMForm( &mouse.gr_mof, &mf_viseur.gr_mof ) ;
                         set_mform(&mouse) ;
                         GWMenuCheck( M_OGOMME ) ;
                         break ;
      case LOUPE       : mouse.gr_monumber = USER_DEF ;
                         if (!config.souris_viseur) CopyMForm( &mouse.gr_mof, &mf_loupe.gr_mof ) ;
                         else                       CopyMForm( &mouse.gr_mof, &mf_viseur.gr_mof ) ;
                         set_mform(&mouse) ;
                         GWMenuCheck( M_OLOUPE ) ;
                         break ;
      case PEINTURE    : mouse.gr_monumber = USER_DEF ;
                         if (!config.souris_viseur) CopyMForm( &mouse.gr_mof, &mf_peint.gr_mof ) ;
                         else                       CopyMForm( &mouse.gr_mof, &mf_viseur.gr_mof ) ;
                         set_mform(&mouse) ;
                         GWMenuCheck( M_OPEINTURE ) ;
                         break ;
      case BLOC       :  if ( !config.souris_viseur ) mouse.gr_monumber = POINT_HAND ;
                         else
                         {
                           mouse.gr_monumber = USER_DEF ;
                           CopyMForm( &mouse.gr_mof, &mf_viseur.gr_mof ) ;
                         }
                         set_mform( &mouse ) ;
                         GWMenuCheck( M_OBLOC ) ;
                         break ;
      case PALETTE    :  CopyMouseDef( &save, &mouse ) ;
                         mouse.gr_monumber = USER_DEF ;
                         CopyMForm( &mouse.gr_mof, &mf_pencil.gr_mof ) ;
                         set_mform(&mouse) ;
                         traite_palette( gwnd ) ;
                         deselect(adr_icones, num) ;
                         select(adr_icones, i) ;
                         memzero(flag, 1+NB_TOOLS) ;
                         flag[i] = flag[num] = flag[PALETTE] = 1 ;
                         VToolBarUpdate( flag ) ;
                         CopyMouseDef( &mouse, &save ) ;
                         set_mform(&mouse) ;
                         break ;
	    case TRACE     :  GWMenuCheck( M_OTRACE ) ;
	                      set_goodmouse() ;
	                      break ;
	    case POINT      : GWMenuCheck( M_OPOINT ) ;
	                      set_goodmouse() ;
	                      break ;
	    case KLIGNE    :  GWMenuCheck( M_OKLIGNE ) ;
	                      set_goodmouse() ;
	                      break ;
	    case LIGNE      : GWMenuCheck( M_OLIGNE ) ;
	                      set_goodmouse() ;
	                      break ;
	    case CERCLE     : GWMenuCheck( M_OCERCLE ) ;
	                      set_goodmouse() ;
	                      break ;
	    case DISQUE     : GWMenuCheck( M_ODISQUE ) ;
	                      set_goodmouse() ;
	                      break ;
	    case RECTANGLE  : GWMenuCheck( M_ORECTANGLE ) ;
	                      set_goodmouse() ;
	                      break ;
	    case PLAN       : GWMenuCheck( M_OPLAN ) ;
	                      set_goodmouse() ;
	                      break ;
	    case TEXTE      : GWMenuCheck( M_OTEXTE ) ;
	                      set_goodmouse() ;
	                      break ;
	    case MAIN       : if (!config.souris_viseur)
                          mouse.gr_monumber = FLAT_HAND ;
                        else
                        {
                          mouse.gr_monumber = USER_DEF ;
                          CopyMForm( &mouse.gr_mof, &mf_viseur.gr_mof ) ;
                        }
                        set_mform(&mouse) ;
                        GWMenuCheck( M_OMAIN ) ;
                        break ;
      case PINCEAU    : mouse.gr_monumber = USER_DEF ;
                        CopyMForm( &mouse.gr_mof, &mf_pencil.gr_mof ) ;
                        GWMenuCheck( M_OPINCEAU ) ;
                        set_mform(&mouse) ;
                        break ;
      case AEROSOL    : mouse.gr_monumber = USER_DEF ;
                        GWMenuCheck( M_OAEROSOL ) ;
                        if (!config.souris_viseur) CopyMForm( &mouse.gr_mof, &mf_aerosol.gr_mof ) ;
                        else                       CopyMForm( &mouse.gr_mof, &mf_viseur.gr_mof ) ;
                        set_mform(&mouse) ;
                        break ;
	    case TABLEAU    : GWMenuCheck( M_OTABLEAU ) ;
	                      set_goodmouse() ;
	                      break ;
      case ZOOM_VUE   : mouse_busy() ;
                        traite_move( gwnd ) ;
                        deselect(adr_icones, num) ;
                        select(adr_icones, i) ;
                        memzero(flag, 1+NB_TOOLS) ;
                        flag[i] = flag[ZOOM_BOX] = flag[PALETTE] = 1 ;
                        VToolBarUpdate( flag ) ;
                        mouse_restore() ;
                        break ;
     default          : set_goodmouse() ;
	                      break ;
    }
  }
}

void traite_espace(GEM_WINDOW* gwnd)
{
  VXIMAGE*  vimage ;
  MOUSE_DEF save ;
  EVENT     ev ;
  int       i ;
  int       dummy, bouton ;
  char      flag[1+NB_TOOLS];

  if ( gwnd == NULL ) return ;
  vimage = (VXIMAGE*) gwnd->Extension ;
  if ( vimage->zoom_level != 1 )
  {
    form_stop( MsgTxtGetMsg(hMsg, MSG_NOTOOLZ) ) ;
    return ;
  }

  CopyMouseDef( &save, &mouse ) ;

  for (i = 1; i <= NB_TOOLS; i++)
    if (selected(adr_icones, i)) break ; 

  if ((i == 1+NB_TOOLS) || (i == MAIN)) return ;
  deselect(adr_icones, i) ;
  if (!config.souris_viseur) mouse.gr_monumber = FLAT_HAND ;
  else
  {
    mouse.gr_monumber = USER_DEF ;
    CopyMForm( &mouse.gr_mof, &mf_viseur.gr_mof ) ;
  }
  set_mform(&mouse) ;
  select(adr_icones, MAIN) ;
  memzero(flag, 1+NB_TOOLS) ; 
  flag[i] = flag[MAIN] = 1 ;
  VToolBarUpdate( flag ) ;

  do
  {
    graf_mkstate(&dummy, &dummy, &bouton, &dummy) ;
  }
  while (bouton != 1) ;
  traite_main( gwnd ) ;
  ev.ev_mflags     = MU_KEYBD | MU_TIMER ;
  ev.ev_mtlocount  = 200 ;
  ev.ev_mthicount  = 0 ;
  while (EvntMulti(&ev) != MU_TIMER) ;
  select(adr_icones, i) ;
  deselect(adr_icones, MAIN) ;
  VToolBarUpdate( flag ) ;
  set_mform(&save) ;
}

void select_zoom(GEM_WINDOW* gwnd, int mzoom)
{
  VXIMAGE* vimage ;
  int      mx, my, dummy ;
  int      last ;

  if ( gwnd == NULL ) return ;
  vimage = (VXIMAGE*) gwnd->Extension ;

  last = vimage->zoom_level ;
  switch(mzoom)
  {
    case M_Z11  : vimage->zoom_level = 1 ;
                  break ;
    case M_Z21  : vimage->zoom_level = 2 ;
                  break ;
    case M_Z41  : vimage->zoom_level = 4 ;
                  break ;
    case M_Z81  : vimage->zoom_level = 8 ;
                  break ;
    case M_Z161 : vimage->zoom_level = 16 ;
                  break ;
  }

  graf_mkstate( &mx, &my, &dummy, &dummy ) ;
  traite_loupe( gwnd, last, mx, my ) ;
}

static void traite_snapshot_key(int key)
{
  if ( key == ASCIICODE(config.snap_keyviaconfig) )   snap( 1 ) ;
  else if ( key == ASCIICODE(config.snap_keydirect) ) snap( 0 ) ;
}

int OnAppKeyPressed(int key)
{
  int code = GW_EVTCONTINUEROUTING ;
  int obj = -1 ;

  switch( key )
  {
    case NOUVEAU     : traite_nouveau() ;
                       break ;

    case OUVRIR      : traite_ouvre(0) ;
                       break ;

    case IMPRIMER    : traite_imprime( NULL ) ;
                       break ;

    case QUITTER     : code = GW_EVTQUITAPP ;
                       break ;

    case PGENERAL    : traite_pgeneral() ;
                       break ;

    case PTEXTE      : traite_ptexte() ;
                       break ;

    case PFORMES     : traite_pformes() ;
                       break ;

    case PGOMME      : traite_pgomme() ;
                       break ;

    case STD_PAL     : traite_tab( NULL ) ;
                       break ;

    case PREF        : traite_preferences() ;
                       set_goodmouse() ;
                       break ;

    case F1          : obj = POINT ;
                       break ;

    case LSF1        : 
    case RSF1        : obj = TRACE ;
                       break ;

    case F2          : obj = PINCEAU ;
                       break ;

    case LSF2       :
    case RSF2       :  obj = AEROSOL ;
                       break ;

    case F3          : obj = KLIGNE ;
                       break ;

    case LSF3        :
    case RSF3        : obj = LIGNE ;
                       break ;

    case F4          : obj = CERCLE ;
                       break ;

    case LSF4        :
    case RSF4        : obj = RECTANGLE ;
                       break ;

    case F5         :  obj = DISQUE ;
                       break ;

    case LSF5        :
    case RSF5        : obj = PLAN ;
                       break ;

    case F6          : obj = PEINTURE ;
                       break ;

    case LSF6        :
    case RSF6        : obj = GOMME ;
                       break ;

    case F7         :  obj = TEXTE ;
                       break ;

    case LSF7        :
    case RSF7        : obj = LOUPE ;
                       break ;

    case F8          : obj = TEMPERATURE ;
                       break ;

    case LSF8        :
    case RSF8        : obj = TABLEAU ;
                       break ;

    case F9         :  obj = MAIN ;
                       break ;

    case LSF9        :
    case RSF9        : obj = BLOC ;
                       break ;

    case F10        :  obj = PALETTE ;
                       break ;
  }

  if ( obj != -1 )
  {
    int xt, yt, dummy ;
    int mx, my ;

    WndToolBar->GetWorkXYWH( WndToolBar, &xt, &yt, &dummy, &dummy ) ;
    graf_mkstate( &mx, &my, &dummy, &dummy ) ;
    mx = xt+adr_icones[obj].ob_x+8 ;
    my = yt+adr_icones[obj].ob_y+8 ;
    select_icon( mx, my ) ;
  }

  traite_snapshot_key( key ) ;

  return code ;
}

void make_undo_item(UNDO_DEF *undo)
{
  int  i ;
  char buffer[80] ;
  char *pt, *pt1 ;

  strcpy( buffer, MsgTxtGetMsg(hMsg, MSG_UNDO)  ) ;
  if ( undo )
  {
    i = undo->mitem ;
    if ( i )
    {
      pt = strchr( buffer+3, ' ' ) ;
      if ( pt )
      {
        pt++ ;
        pt1 = MsgTxtGetMsg(hMsg, i+MSG_UNDO)  ;
        while ( *pt1 ) *pt++ = *pt1++ ;
      }
    }
  }
  menu_text( GemApp.Menu, M_UNDO, buffer ) ;
}

void OnAppUpdateGUI(void)
{
  GEM_WINDOW* gwnd = GemApp.CurrentGemWindow ;
  OBJECT*     menu ;
  VXIMAGE*    vimage = NULL ;
  UNDO_DEF*   undo = NULL ;

  if ( gwnd && (strcmp( gwnd->ClassName, VIMG_CLASSNAME ) == 0) )
  {
    vimage = (VXIMAGE*) gwnd->Extension ;
    undo   = GetUndoBuffer( vimage ) ;
  }

  menu = GemApp.Menu ;
  GWSetMenuStatus( 1 ) ;
  menu_ienable( menu, M_BARATIN, 1 ) ; /* Cas special : sinon A propos de ... reste grise ??? */
  make_undo_item( undo ) ;
  if ( vimage )
  {
    int check_zoom = 0 ;

    if ( undo == NULL )
    {
      menu[M_UNDO].ob_state |= DISABLED ;
      if ( vimage->Redo == NULL ) menu[M_CLEARBUF].ob_state |= DISABLED ;
    }

	  if ( vimage->Redo == NULL ) menu[M_REFAIRE].ob_state |= DISABLED ;

    switch( vimage->zoom_level )
    {
      case 1  : if ( ( menu[M_Z11].ob_state & CHECKED ) == 0 ) check_zoom = M_Z11 ;
                break ;
      case 2  : if ( ( menu[M_Z21].ob_state & CHECKED ) == 0 ) check_zoom = M_Z21 ;
                break ;
      case 4  : if ( ( menu[M_Z41].ob_state & CHECKED ) == 0 ) check_zoom = M_Z41 ;
                break ;
      case 8  : if ( ( menu[M_Z81].ob_state & CHECKED ) == 0 ) check_zoom = M_Z81 ;
                break ;
      case 16 : if ( ( menu[M_Z161].ob_state & CHECKED ) == 0 ) check_zoom = M_Z161 ;
                break ;
    }
    if ( check_zoom != 0 )
    {
      GWMenuUncheck( M_Z11 ) ;
      GWMenuUncheck( M_Z21 ) ;
      GWMenuUncheck( M_Z41 ) ;
      GWMenuUncheck( M_Z81 ) ;
      GWMenuUncheck( M_Z161 ) ;
      GWMenuCheck( check_zoom ) ;
    }

    if ( !GWIsWindowValid( vclip.gwindow ) )
	  {
	    menu[M_COUPE].ob_state   |= DISABLED ;
	    menu[M_COPIE].ob_state   |= DISABLED ;
	    menu[M_EFFACE].ob_state  |= DISABLED ;
	    menu[M_NEGATIF].ob_state |= DISABLED ;
	  }

	  if ( nb_colors < 256 ) menu[M_CONVOLUTION].ob_state |= DISABLED ; /*menu_ienable( GemApp.Menu, M_CONVOLUTION, 0 ) ;*/
	  if ( Truecolor )       menu[M_PALETTE].ob_state |= DISABLED ; /*menu_ienable( GemApp.Menu, M_PALETTE, 0 ) ;*/
  }
  else
  {
    char items[] = { M_UNDO, M_CLEARBUF, M_REFAIRE, M_COUPE, M_COLLE, M_COPIE, M_EFFACE, M_NEGATIF, M_SELECTIONNE,
                     M_PLEIN, M_PLEINECRAN, M_INFO, M_CONVERSION, M_Z11, M_Z21, M_Z41, M_Z81, M_Z161, M_SYMETRIE, 
                     M_LDV, M_MKPALETTE, M_TAILLE, M_CONVOLUTION, M_HISTO } ;
    short i ;

    if ( gwnd && (strcmp( gwnd->ClassName, HEXDUMP_CLASSNAME ) == 0) )
    {
	    menu[M_FERME].ob_state  &= ~DISABLED ;
	    menu[M_SAUVE].ob_state  &= ~DISABLED ;
	    menu[M_SAUVES].ob_state &= ~DISABLED ;
    }
    else
    {
	    menu[M_FERME].ob_state  |= DISABLED ;
	    menu[M_SAUVE].ob_state  |= DISABLED ;
	    menu[M_SAUVES].ob_state |= DISABLED ;
    }

    for ( i = 0; i < ARRAY_SIZE(items); i++ )
      menu[items[i]].ob_state |= DISABLED ;
  }

  if ( config.color_protect == 0 ) GWMenuUncheck( M_PALETTE ) ;

#ifdef __NO_PRINT
  menu[M_IMPRIME].ob_state |= DISABLED ;
#endif  
#ifdef __NO_BATCH_CONV
  menu[M_BATCHCONV].ob_state |= DISABLED ;
#endif  
}

int OnAppMenuSelected(int t_id, int m_id)
{
  char m_otools[] = { M_OPOINT, M_OTRACE, M_OLIGNE, M_OKLIGNE, M_ORECTANGLE, M_OPLAN, M_OCERCLE, M_ODISQUE, M_OPEINTURE, M_OPALETTE, M_OGOMME, M_OTEXTE, M_OPINCEAU, M_OAEROSOL, M_OTABLEAU, M_OTEMP, M_OLOUPE, M_OMAIN, M_OBLOC } ;
  int  keys[] =     {       F1,     LSF1,     LSF3,        F3,         LSF4,    LSF5,        F4,        F5,          F6,        F10,     LSF6,       F7,         F2,       LSF2,       LSF8,      F8,     LSF7,      F9,    LSF9 } ;
  int  code = GW_EVTCONTINUEROUTING ;
  int  i ;

  for ( i = 0; i < ARRAY_SIZE(m_otools); i++ )
    if ( m_id == m_otools[i] ) { OnAppKeyPressed( keys[i] ) ; break ; } ;

  if ( i == ARRAY_SIZE(m_otools) )
  {
    /* Not found, try others */
    switch( m_id )
    {
      case M_BARATIN   : display_baratin( high_res ) ;
                         break ;
      case M_NOUVEAU   : traite_nouveau() ;
                         break ;
      case M_OUVRE     : traite_ouvre( 0 ) ;
                         break ;
      case M_OUVRES    : traite_ouvre( 1 ) ;
                         break ;
      case M_BATCHCONV : traite_batch_conversion( NULL ) ;
                         break ;
      case M_IMPRIME   : traite_imprime( NULL ) ;
                         break ;
      case M_QUITTE    : code = GW_EVTQUITAPP ;
                         break ;
      case M_NCOLLE    : traite_ncolle( NULL ) ;
                         break ;
      case M_CATALOG   : traite_album( NULL ) ;
                         break ;
      case M_PGENERAL  : traite_pgeneral() ;
                         break ;
      case M_PTEXTE    : traite_ptexte() ;
                         break ;
      case M_PFORMES   : traite_pformes() ;
                         break ;
      case M_PGOMME    : traite_pgomme() ;
                         break ;
      case M_PALETTE   : traite_tab( NULL ) ;
                         break ;
      case M_PREF      : traite_preferences() ;
                         set_goodmouse() ;
                         break ;
    }
  }

/*
	case M_OPOINT       : OnAppKeyPressed( F1 ) ;
                        break ;
	case M_OTRACE       : OnAppKeyPressed( LSF1 ) ;
                        break ;
	case M_OLIGNE       : OnAppKeyPressed( LSF3 ) ;
                        break ;
	case M_OKLIGNE      : OnAppKeyPressed( F3 ) ;
                        break ;
	case M_ORECTANGLE   : OnAppKeyPressed( LSF4 ) ;
                        break ;
	case M_OPLAN        : OnAppKeyPressed( LSF5 ) ;
                        break ;
	case M_OCERCLE      : OnAppKeyPressed( F4 ) ;
	                      break ;
	case M_ODISQUE      : OnAppKeyPressed( F5 ) ;
	                      break ;
	case M_OPEINTURE    : OnAppKeyPressed( F6 ) ;
	                      break ;
	case M_OPALETTE     : OnAppKeyPressed( F10 ) ;
	                      break ;
	case M_OGOMME       : OnAppKeyPressed( LSF6 ) ;
	                      break ;
	case M_OTEXTE       : OnAppKeyPressed( F7 ) ;
	                      break ;
	case M_OPINCEAU     : OnAppKeyPressed( F2 ) ;
	                      break ;
	case M_OAEROSOL     : OnAppKeyPressed( LSF2 ) ;
	                      break ;
	case M_OTABLEAU     : OnAppKeyPressed( LSF8 ) ;
	                      break ;
	case M_OTEMP        : OnAppKeyPressed( F8 ) ;
	                      break ;
	case M_OLOUPE       : OnAppKeyPressed( LSF7 ) ;
	                      break ;
	case M_OMAIN        : OnAppKeyPressed( F9 ) ;
	                      break ;
	case M_OBLOC        : OnAppKeyPressed( LSF9 ) ;
	                      break ;
  }
*/
  GWOnMenuSelected( t_id, m_id ) ;

  return code ;
}

void param_icon(int mx, int my)
{
  int ob_num ;
      
  ob_num = objc_find( adr_icones, 0, 200, mx, my ) ;
  switch( ob_num )
  {
    case PLAN      :
    case TRACE     :
    case POINT     :
    case LIGNE     :
    case KLIGNE    :
    case PEINTURE  :
    case RECTANGLE :
    case TABLEAU   :
    case DISQUE    : traite_pformes() ;
                     break ;
    case PINCEAU   : traite_ppinceau() ;
                     break ;
    case TEXTE     : traite_ptexte() ;
                     break ;
    case GOMME     : traite_pgomme() ;
                     break ;
    case AEROSOL   : traite_paerosol() ;
                     break ;
    case BLOC      : traite_pbloc() ;
                     break ;
  }
}

int OnVImgTopped(GEM_WINDOW* wnd)
{
  GWOnTopped( wnd ) ;
  set_imgpalette( wnd->Extension ) ;

  return 0 ;
}

void iconify_picture(GEM_WINDOW* wnd)
{
  VXIMAGE* vimage ;
  MFDB     temp ;
  float    pcx, pc ;
  int      xy[8] ;
  int      w, h, dummy ;
  int      nw ;

  vimage = (VXIMAGE*) wnd->Extension ;
  wnd->GetWorkXYWH( wnd, &dummy, &dummy, &w, &h ) ;
  ZeroMFDB( &wnd->window_icon ) ;
  wnd->window_icon.fd_addr = img_alloc( w, h, screen.fd_nplanes ) ;
  if ( wnd->window_icon.fd_addr == NULL ) return ;
  wnd->window_icon.fd_w = w ;
  wnd->window_icon.fd_h = h ;
  wnd->window_icon.fd_wdwidth = w2wdwidth( w ) ;
  wnd->window_icon.fd_nplanes = screen.fd_nplanes ;

  img_raz( &wnd->window_icon ) ;
  pcx = (float)w/(float)vimage->inf_img.mfdb.fd_w ;
  pc  = (float)h/(float)vimage->inf_img.mfdb.fd_h ;
  if (pc > pcx) pc = pcx ;
  /* Rduit la taille de l'image  une icone */
  temp.fd_addr = NULL ;
  temp.fd_w    = (int) ( 0.5 + (float)vimage->inf_img.mfdb.fd_w * pc ) ;
  temp.fd_h    = (int) ( 0.5 + (float)vimage->inf_img.mfdb.fd_h * pc ) ;
  RasterZoom( &vimage->inf_img.mfdb, &temp, NULL ) ;
  if ( temp.fd_addr == NULL )
  {
    Xfree( wnd->window_icon.fd_addr ) ;
    ZeroMFDB( &wnd->window_icon ) ;
    return ;
  }

  nw = (int) ((float)vimage->inf_img.mfdb.fd_w*pc) ;
  if (nw > w) nw = w ;
  xy[0] = xy[1] = 0 ;
  xy[2] = nw-1 ;
  xy[3] = temp.fd_h-1 ;
  xy[4] = (w-nw) >> 1 ;
  xy[5] = (h-temp.fd_h) >> 1 ;
  xy[6] = xy[4]+nw-1 ;
  xy[7] = xy[5]+temp.fd_h-1 ;
  vro_cpyfm(handle, S_ONLY, xy, &temp,&wnd->window_icon ) ;
  Xfree( temp.fd_addr ) ;
}

int GetRVBForVIMG(GEM_WINDOW* wnd, int mx, int my, int rvb[3])
{
  VXIMAGE* vimage = wnd->Extension ;
  int      index = -1 ;
  int      xpos, ypos ;
  int      xi, yi, dummy ;

  if ( Truecolor )
  {
    xpos = vimage->x1 + mx / vimage->zoom_level ;
    ypos = vimage->y1 + my / vimage->zoom_level ;
    GetPixel( &vimage->inf_img.mfdb, xpos, ypos, rvb ) ;
  }
  else
  {
    wnd->GetWorkXYWH( wnd, &xi, &yi, &dummy, &dummy ) ;
    GWHideCursor( 0 ) ;
    v_get_pixel( handle, xi + mx, yi + my, &dummy, &index ) ;
    GWShowCursor( 0 ) ;
    vq_color( handle, index, 1, rvb ) ;
  }

  return index ;
}

void draw_selection(GEM_WINDOW* wnd, int xywh[4], int xi, int yi)
{
  VXIMAGE* vimage = (VXIMAGE*) wnd->Extension ;
  int      cxy[4] ;
  int      posx, posy ;
  int      x1, y1, x2, y2 ;

  if ( vimage->zoom_level != 1 ) return ;

  posx = vimage->x1 ;
  posy = vimage->y1 ;
  vswr_mode( handle, MD_REPLACE ) ;
  vsl_color( handle, 1 ) ;
  vsl_ends( handle, SQUARE, SQUARE ) ;
  vsl_width( handle, 1 ) ;
  vsl_type( handle, DOT ) ;

  x1 = xi+vclip.x1-posx ; y1 = yi+vclip.y1-posy ;
  x2 = xi+vclip.x2-posx ; y2 = yi+vclip.y2-posy ;
  cxy[0] = xywh[0] ; cxy[1] = xywh[1] ;
  cxy[2] = cxy[0] + xywh[2] - 1 ;
  cxy[3] = cxy[1] + xywh[3] - 1 ;
  vs_clip( handle, 1, cxy ) ;
  rectangle( x1, y1, x2, y2 ) ;
  vs_clip( handle, 0, cxy ) ;
}

void paint_zone(int* xy)
{
  int index ;

  if ( !Truecolor ) index = 0 ;
  else
  {
    int rgb[3] = { 1000, 1000, 1000 } ;

    index = 19 ;
    vs_color( handle, index, rgb ) ;
  }
  vswr_mode( handle, MD_REPLACE ) ;
  vsf_color( handle, index ) ;
  vsf_interior( handle, FIS_SOLID ) ;
  vsf_perimeter( handle, 0 ) ;
  vr_recfl( handle, xy ) ;
}

void OnVImgDraw(GEM_WINDOW* wnd, int xywh[4])
{
  VXIMAGE* vimage ;
  int      w, h ;
  int      off_x, off_y, xi, yi, dummy ;
  int      xy[8] ;
  int      posx, posy ;
  int      must_paint = 0 ;

  vimage = (VXIMAGE*) wnd->Extension ;

  if ( vimage->zoom_level == 1 )
  {
    posx = vimage->x1 ;
    posy = vimage->y1 ;
  }
  else
    posx = posy = 0 ;

  wnd->GetWorkXYWH( wnd, &xi, &yi, &dummy, &dummy ) ;
  off_x = xywh[0] - xi ; off_y = xywh[1] - yi ;
  w     = xywh[2] ;      h     = xywh[3] ;

  /* Sanity checks for painting */
  xy[0] = off_x + posx ;   xy[1] = off_y + posy ;
  if ( vimage->zoom_level == 1 )
  {
    if ( xy[0] + w - 1 > vimage->inf_img.mfdb.fd_w )
    {
      w          = 1 + vimage->inf_img.mfdb.fd_w - xy[0] ;
      must_paint = 1 ;
    }

    if ( xy[1] + h - 1 > vimage->inf_img.mfdb.fd_h )
    {
      h          = 1 + vimage->inf_img.mfdb.fd_h - xy[1] ;
      must_paint = 1 ;
    }
  }

  if ( must_paint )
  {
    int xys[4] ;

    xys[0] = xywh[0] ;
    xys[1] = xywh[1] ;
    xys[2] = xywh[0] + xywh[2] - 1 ;
    xys[3] = xywh[1] + xywh[3] - 1 ;
    paint_zone( xys ) ;
  }

  xy[2] = xy[0] + w - 1 ;   xy[3] = xy[1] + h - 1 ;
  xy[4] = xywh[0] ;         xy[5] = xywh[1] ;
  xy[6] = xywh[0] + w - 1 ; xy[7] = xywh[1] + h - 1 ;

  /* Debug crash on Magic when loading demoscene1.jpg */
/*  LogMFDB(&vimage->inf_img.mfdb) ;
  LoggingDo(LL_GW2, "+OnVImgDraw(%d)xy=%d,%d,%d,%d,%d,%d,%d,%d", vimage->zoom_level, xy[0], xy[1], xy[2], xy[3], xy[4], xy[5], xy[6], xy[7] ) ;
  LogMFDB(&screen) ;*/
  vro_cpyfm( handle, S_ONLY, xy, (vimage->zoom_level == 1) ? &vimage->inf_img.mfdb:&vimage->zoom, &screen ) ;

  if ( vclip.gwindow == wnd ) draw_selection( wnd, xywh, xi, yi ) ;
}
/*
void OnVImgDraw(GEM_WINDOW* wnd, int xywh[4])
{
  VXIMAGE *vimage ;
  int     w, h ;
  int     off_x, off_y, xi, yi, dummy ;
  int     xy[8] ;
  int     posx, posy ;

  vimage = (VXIMAGE *) wnd->Extension ;

  if ( vimage->zoom_level == 1 )
  {
    posx = vimage->x1 ;
    posy = vimage->y1 ;
  }
  else
    posx = posy = 0 ;

  w = xywh[2] ;
  h = xywh[3] ;
  xy[4] = xywh[0] ;         xy[5] = xywh[1] ;
  xy[6] = xywh[0] + w - 1 ; xy[7] = xywh[1] + h - 1 ;

  wnd->GetWorkXYWH( wnd, &xi, &yi, &dummy, &dummy ) ;
  off_x = xywh[0] - xi ;
  off_y = xywh[1] - yi ;
  xy[0] = off_x + posx ; xy[1] = off_y + posy ;
  xy[2] = xy[0] + w - 1 ;
  xy[3] = xy[1] + h - 1 ;
  
  if ( vimage->zoom_level == 1 ) vro_cpyfm( handle, S_ONLY, xy, &vimage->inf_img.mfdb, &screen ) ;
  else                           vro_cpyfm( handle, S_ONLY, xy, &vimage->zoom, &screen ) ;

  if ( vclip.gwindow == wnd ) draw_selection( wnd, xywh, xi, yi ) ;
}
*/

void FreeRelatedMemory(GEM_WINDOW* gwnd)
{
  traite_effbuffers( gwnd ) ;

  if ( clipboard.gwindow == gwnd ) FreeClipboard() ;

  if ( vclip.gwindow == gwnd )
  {
    vclip.gwindow = NULL ;
    vclip.x1  = 0 ; vclip.y1  = 0 ;
    vclip.x2  = 0 ; vclip.y2  = 0 ;
  }
}

int OnVImgClose(GEM_WINDOW* gwnd)
{
  VXIMAGE* vimage = (VXIMAGE*) gwnd->Extension ;
  int      code = GW_EVTCONTINUEROUTING ;

  if ( vimage->modif )
  {
    if ( form_interrogation(1, MsgTxtGetMsg(hMsg, MSG_IMAGECHANGE)  ) == 1 )
      traite_sauves( gwnd, 0 ) ;
  }

  if ( vimage->zoom.fd_addr ) Xfree( vimage->zoom.fd_addr ) ;
  free_info( &vimage->inf_img ) ;
  FreeRelatedMemory( gwnd ) ;
  GWOnClose( gwnd ) ;

  if ( GWGetWindowByClass( VIMG_CLASSNAME ) == NULL )
  {
    /* No more window open */
    if ( config.quitte )  code = GW_EVTQUITAPP ;      /* Quit */
    else if ( !Truecolor) set_tospalette( NULL, 0 ) ; /* Restore original palette */
  }

  return code ;
}

int traite_thickness(int key)
{
  char  key_thick[] = { THICKNESS1, THICKNESS2, THICKNESS3, THICKNESS4, THICKNESS5,
                        THICKNESS6, THICKNESS7, THICKNESS8, THICKNESS9 } ;
  short i ;
  int   changed = 0 ;

  for ( i = 0; i < ARRAY_SIZE(key_thick); i++)
  {
    if ( key_thick[i] == key )
    {
      changed = config.thickness[i] ;
      break ;
    }
  }

  if ( changed )
  {
    LoggingDo(LL_DEBUG, "Thickness changed to %d", changed) ;
    mtype.height = ltype.width = pencil.height = changed ;
  }

  return changed ;
}

int OnVImgKeyPressed(GEM_WINDOW* gwnd, int key)
{
  VXIMAGE* vimage = (VXIMAGE*) gwnd->Extension ;
  int      xt, yt, mx, my, dummy ;
  int      code = GW_EVTSTOPROUTING ;

  switch( key )
  {
    case FERMER      : gwnd->OnClose( gwnd ) ;
                       break ;

	  case SAUVE       : traite_sauves( gwnd, 0 ) ;
                       break ;

    case SAUVES    :   traite_sauves( gwnd, 1 ) ;
                       break ;

    case IMPRIMER :    traite_imprime( gwnd ) ;
                       break ;

    case UNDO      :   undo_restore( gwnd ) ;
                       break ;

    case COUPER      : traite_coupe() ;
                       break ;

	case COPIER      : traite_copie() ;
	                   break ;

	case COLLER      : traite_colle( gwnd, -1, -1 ) ;
	                   break ;

  case EFFACE      : traite_efface() ;
                       break ;

  case NEGATIF     : traite_negatif( gwnd ) ;
                       break ;

  case S_ALL       : traite_selectall( gwnd ) ;
                       break ;

  case STD_PAL     : traite_tab( gwnd ) ;
	                   break ;

  case PLEINE_PAGE : traite_fullscreen( gwnd ) ;
	                   break ;

  case INFOS       : traite_info( gwnd ) ;
	                   break ;

	case CONVERSION  : traite_conversion( gwnd ) ;
	                   break ;

	case ESPACE      : traite_espace( gwnd ) ;
	                   break ;

  case ZOOM1       : select_zoom( gwnd, M_Z11) ;
                       break ;

  case ZOOM2       : select_zoom( gwnd, M_Z21) ;
                       break ;

  case ZOOM4       : select_zoom( gwnd, M_Z41) ;
                       break ;

  case ZOOM8       : select_zoom( gwnd, M_Z81) ;
                       break ;

  case ZOOM16      : select_zoom( gwnd, M_Z161) ;
                       break ;

  case ZOOMPLUS    : vimage->zoom_level *= 2 ;
                       if ( vimage->zoom_level > 16 )
                       {
                         vimage->zoom_level = 16 ;
                         return GW_EVTSTOPROUTING ;
                       }
                       graf_mkstate( &mx, &my, &dummy, &dummy ) ;
                       traite_loupe( gwnd, vimage->zoom_level, mx, my ) ;
                       break ;

    case ZOOMMOINS   : vimage->zoom_level >>= 1 ;
                       if ( vimage->zoom_level == 0 )
                       {
                         vimage->zoom_level = 1 ;
                         return GW_EVTSTOPROUTING ;
                       }
                       graf_mkstate( &mx, &my, &dummy, &dummy ) ;
                       traite_loupe( gwnd, vimage->zoom_level, mx, my ) ;
                       break ;

	  case SYMETRIE    : traite_symetrie( gwnd ) ;
	                     break ;

	  case TAILLE      : traite_taille( gwnd ) ;
	                     break ;

	  case LDV         : traite_ldv( gwnd ) ;
	                     break ;

	  case FILTRE      : traite_convolution( gwnd ) ;
	                     break ;

    case DOPALETTE   : traite_makepalette( gwnd ) ;
                       break ;

    case PIPETTE    :  WndToolBar->GetWorkXYWH( WndToolBar, &xt, &yt, &dummy, &dummy ) ;
                       mx = xt+adr_icones[TEMPERATURE].ob_x+8 ;
	                     my = yt+adr_icones[TEMPERATURE].ob_y+8 ;
	                     select_icon( mx, my ) ;
                       break ;

    default :          code = GW_EVTCONTINUEROUTING ;
                       break ;
  }

  traite_thickness( key ) ;

  return code ;
}

int OnVImgMenuSelected(GEM_WINDOW* wnd, int t_id, int m_id)
{
  int code = GW_EVTSTOPROUTING ;

  switch( m_id )
  {
	case M_FERME        : wnd->OnClose( wnd ) ;
                        break ;

	case M_SAUVE        : traite_sauves( wnd, 0 ) ;
	                      break ;

	case M_SAUVES       : traite_sauves( wnd, 1 ) ;
	                      break ;

  case M_IMPRIME      : traite_imprime( wnd ) ;
                        break ;

	case M_UNDO         : undo_restore( wnd ) ;
	                      break ;

	case M_REFAIRE      : traite_refaire( wnd ) ;
	                      break ;

	case M_CLEARBUF     : traite_effbuffers( wnd ) ;
	                      break ;

	case M_COUPE        : traite_coupe() ;
	                      break ;

	case M_COPIE        : traite_copie() ;
	                      break ;

	case M_COLLE        : traite_colle( wnd, -1, -1 ) ;
	                      break ;

  case M_EFFACE       : traite_efface() ;
                        break ;

  case M_NEGATIF       : traite_negatif( wnd ) ;
                         break ;

  case M_SELECTIONNE  : traite_selectall( wnd ) ;
                        break ;

	case M_PLEIN        : traite_fullscreen( wnd ) ;
	                      break ;

	case M_PLEINECRAN   : traite_pleinecran( wnd ) ;
	                      break ;

	case M_INFO         : traite_info( wnd ) ;
	                      break ;

	case M_CONVERSION   : traite_conversion( wnd ) ;
	                      break ;

	case M_Z11          :
	case M_Z21          :
	case M_Z41          :
	case M_Z81          :
	case M_Z161         : select_zoom( wnd, m_id ) ;
	                      break ;

	case M_SYMETRIE     : traite_symetrie( wnd ) ;
	                      break ;

	case M_TAILLE       : traite_taille( wnd ) ;
	                      break ;

	case M_LDV          : traite_ldv( wnd ) ;
	                      break ;

	case M_CONVOLUTION  : traite_convolution( wnd ) ;
	                      break ;

	case M_PALETTE      : traite_tab( wnd ) ;
	                      break ;

    case M_MKPALETTE    : traite_makepalette( wnd ) ;
                          break ;

    case M_HISTO        : traite_histo( wnd ) ;
                          break ;

    default :             code = GW_EVTCONTINUEROUTING ;
                          break ;
  }

  GWOnMenuSelected( t_id, m_id ) ;

  return code ;
}

int OnVImgArrow(GEM_WINDOW* gwnd, int action)
{
  VXIMAGE *vimage = (VXIMAGE*) gwnd->Extension ;
  long    wdiv, hdiv ;
  int     new ;
  int     rc, rl ;
  int     w, h ;
  int     index ;
  int     hsld = 1 ;
 
  for (index = 0; index < 4; index++)
    if ( vimage->zoom_level == zooms[index] ) break ;

  rc = vimage->x1 ;
  rl = vimage->y1 ;
  w  = 1+vimage->x2-vimage->x1 ;
  h  = 1+vimage->y2-vimage->y1 ;
  wdiv = vimage->inf_img.mfdb.fd_w-w ;
  if (wdiv <= 0) wdiv = 1 ;
  hdiv = vimage->inf_img.mfdb.fd_h-h ;
  if (hdiv <= 0) hdiv = 1 ;
  switch(action)
  {
    case WA_UPPAGE : rl -= h ;
                     if (rl < 0) rl = 0 ;
                     new = pm( rl, hdiv ) ;
                     hsld = 0 ;
                     break ;
    case WA_DNPAGE : rl += h ;
                     if (rl+h >= vimage->inf_img.mfdb.fd_h)
                       rl = vimage->inf_img.mfdb.fd_h-h ;
                     new = pm( rl, hdiv ) ;
                     hsld = 0 ;
                     break ;
    case WA_LFPAGE : rc -= w ;
                     if (rc < 0) rc = 0 ; 
                     new = pm( rc, wdiv ) ;
                     break ;
    case WA_RTPAGE : rc += w ;
                     if (rc+w >= vimage->inf_img.mfdb.fd_w)
                       rc = vimage->inf_img.mfdb.fd_w-w ;
                     new = pm( rc, wdiv ) ;
                     break ;
    case WA_UPLINE : rl -= 8 ;
                     if (rl < 0) rl = 0 ;
                     new = pm( rl, hdiv ) ;
                     hsld = 0 ;
                     break ;
    case WA_DNLINE : rl += 8 ;
                     if (rl+h >= vimage->inf_img.mfdb.fd_h)
                       rl = vimage->inf_img.mfdb.fd_h-h ;
                     new = pm( rl, hdiv ) ;
                     hsld = 0 ;
                     break ;
    case WA_LFLINE : rc -= 8 ;
                     if (rc < 0) rc = 0 ;  
                     new = pm( rc, wdiv ) ;
                     break ;
    case WA_RTLINE : rc += 8 ;
                     if (rc+w >= vimage->inf_img.mfdb.fd_w)
                       rc = vimage->inf_img.mfdb.fd_w-w ;
                     new = pm( rc, wdiv ) ;
                     break ;
  }

  vimage->x1 = rc ;
  vimage->y1 = rl ;
  vimage->x2 = rc+w-1 ;
  vimage->y2 = rl+h-1 ;
  if ( vimage->zoom_level == 1 )
  {
    GWRePaint( gwnd ) ;
    if ( hsld ) GWSetHSlider( gwnd, new ) ;
    else        GWSetVSlider( gwnd, new ) ;
  }
  else traite_loupe( gwnd, vimage->zoom_level, -1, -1 ) ;

  update_zview( gwnd ) ;

  return GW_EVTSTOPROUTING ;
}

static void FinishVImgHVSlider(GEM_WINDOW* wnd, int zoom_level)
{
  if ( zoom_level == 1 ) GWRePaint( wnd ) ;
  else                   traite_loupe( wnd, zoom_level, -1, -1 ) ;
  
  update_zview( wnd ) ;
}

void OnVImgHSlider(GEM_WINDOW* wnd, int pos)
{
  VXIMAGE* vimage = (VXIMAGE*) wnd->Extension ;
  int      wxy ;

  wxy        = 1 + vimage->x2 - vimage->x1 ;
  vimage->x1 = pml( pos, vimage->inf_img.mfdb.fd_w-wxy ) ;
  vimage->x2 = vimage->x1 + wxy - 1 ;
  GWOnHSlider( wnd, pos ) ;
  FinishVImgHVSlider( wnd, vimage->zoom_level ) ;
}

void OnVImgVSlider(GEM_WINDOW* wnd, int pos)
{
  VXIMAGE* vimage = (VXIMAGE*) wnd->Extension ;
  int      wxy ;

  wxy    = 1 + vimage->y2 - vimage->y1 ;
  vimage->y1 = pml( pos, vimage->inf_img.mfdb.fd_h-wxy ) ;
  vimage->y2 = vimage->y1 + wxy - 1 ;
  GWOnVSlider( wnd, pos ) ;
  FinishVImgHVSlider( wnd, vimage->zoom_level ) ;
}

void OnVImgSize(GEM_WINDOW* wnd, int xywh[4])
{
  VXIMAGE *vimage = (VXIMAGE*) wnd->Extension ;
  void    *virtuel ;
  int     do_it = 1 ;
  int     rc, rl ;
  int     index ;
  int     wxi, whi ;
  int     xi, yi, wxe, whe ;

  for ( index = 0; index < 4; index++ )
    if ( vimage->zoom_level == zooms[index] ) break ;

  if ( vimage->zoom_level != 1 )
  {
    /* Teste si il y a suffisament de mmoire pour le zoom */
    virtuel = img_alloc( xywh[2], xywh[3], screen.fd_nplanes ) ;
    if ( virtuel == NULL )
    {
      do_it = 0 ;
      form_error(8) ;
    }
    else Xfree( virtuel ) ;
  }
  if ( !do_it ) return ;

  rc = vimage->x1 ;
  rl = vimage->y1 ;
  wind_calc( WC_WORK, wnd->window_kind, xywh[0], xywh[1], xywh[2], xywh[3], &xywh[0], &xywh[1], &xywh[2], &xywh[3]) ;
  wxi = xywh[2]/zooms[index] ;
  whi = xywh[3]/zooms[index] ;
  if (wxi > vimage->inf_img.mfdb.fd_w) wxi = vimage->inf_img.mfdb.fd_w ;
  if (whi > vimage->inf_img.mfdb.fd_h) whi = vimage->inf_img.mfdb.fd_h ;
  if (rc+wxi >= vimage->inf_img.mfdb.fd_w) rc = vimage->inf_img.mfdb.fd_w-wxi ;
  if (rl+whi >= vimage->inf_img.mfdb.fd_h) rl = vimage->inf_img.mfdb.fd_h-whi ;
  vimage->x1 = rc ;
  vimage->y1 = rl ;
  vimage->x2 = rc+wxi-1 ;
  vimage->y2 = rl+whi-1 ;
  if ( vimage->zoom_level != 1 ) traite_loupe( wnd, vimage->zoom_level, -1, -1 ) ;
  else
  {
    wind_calc( WC_BORDER, wnd->window_kind, xywh[0], xywh[1], wxi, whi, &xi, &yi, &wxe, &whe ) ;
    xywh[0] = xi ; xywh[1] = yi ;
    xywh[2] = wxe ; xywh[3] = whe ; /* Pour GWOnSize quand il fera wind_set( .., WF_CURRXYWH */
    xi = pmfr( vimage->x1, vimage->inf_img.mfdb.fd_w-wxi ) ;
    yi = pmfr( vimage->y1, vimage->inf_img.mfdb.fd_h-whi ) ;
    GWSetHSlider( wnd, xi) ;
    GWSetVSlider( wnd, yi) ;
    wxe = pmfr( wxi, vimage->inf_img.mfdb.fd_w ) ;
    whe = pmfr( whi, vimage->inf_img.mfdb.fd_h ) ;
    GWSetHSliderSize( wnd, wxe) ;
    GWSetVSliderSize( wnd, whe) ;
    update_zview( wnd ) ;
    GWOnSize( wnd, xywh ) ;
  }
}

int OnVImgFulled(GEM_WINDOW* wnd)
{
  VXIMAGE *vimage = (VXIMAGE*) wnd->Extension ;
  float   zratioi ;
  int     xywh[4] ;
  int     xi, yi ;
  int     xe, ye, wxe, whe ;
  int     wxee, whee ;
  int     index ;
  int     offset ;

  for (index = 0; index < 4; index++)
    if ( vimage->zoom_level == zooms[index] ) break ;

  wind_calc( WC_WORK, wnd->window_kind, xopen, yopen, wdesk-(xopen-xdesk), hdesk-(yopen-ydesk),
             &xe, &ye, &wxe, &whe ) ;

  zratioi = zratio[index] ;
  /* On s'assure que wxe et whe sont multiples du niveau de zoom */
  wxe  = (int)zratioi*(wxe/(int)zratioi) ;
  whe  = (int)zratioi*(whe/(int)zratioi) ;

  wxee = (int) ((float)wxe/zratioi) ;
  whee = (int) ((float)whe/zratioi) ;
  if (wxee > vimage->inf_img.mfdb.fd_w)
  {
    wxe  = (vimage->inf_img.mfdb.fd_w)*zooms[index] ;
    wxee = (int) ((float)wxe/zratioi) ;
  }
  if (whee > vimage->inf_img.mfdb.fd_h)
  {
    whe  = (vimage->inf_img.mfdb.fd_h)*zooms[index] ;
    whee = (int) ((float)whe/zratioi) ;
  }

  wind_calc( WC_BORDER, wnd->window_kind, xe, ye, wxe, whe, &xywh[0], &xywh[1], &xywh[2], &xywh[3] ) ;
  vimage->x2 = vimage->x1 + wxee - 1 ;
  vimage->y2 = vimage->y1 + whee - 1 ;
  if ( vimage->x2 >= vimage->inf_img.mfdb.fd_w )
  {
    offset = 1 + vimage->x2 - vimage->inf_img.mfdb.fd_w ;

    vimage->x1 -= offset ;
    vimage->x2 -= offset ;
  }
  if ( vimage->y2 >= vimage->inf_img.mfdb.fd_h )
  {
    offset = 1 + vimage->y2 - vimage->inf_img.mfdb.fd_h ;

    vimage->y1 -= offset ;
    vimage->y2 -= offset ;
  }
 
  xi = pmfr( vimage->x1, vimage->inf_img.mfdb.fd_w-wxee ) ;
  yi = pmfr( vimage->y1, vimage->inf_img.mfdb.fd_h-whee ) ;
  GWSetHSlider( wnd, xi ) ;
  GWSetVSlider( wnd, yi ) ;
  wxe = pmfr( wxee, vimage->inf_img.mfdb.fd_w ) ;
  whe = pmfr( whee, vimage->inf_img.mfdb.fd_h ) ;
  GWSetHSliderSize( wnd, wxe ) ;
  GWSetVSliderSize( wnd, whe ) ;
  if ( vimage->zoom_level != 1 )
  {
    traite_loupe( wnd, vimage->zoom_level, -1, -1 ) ;
    wnd->must_be_fulled = 0 ;
  }
  else GWOnFulled( wnd, xywh ) ;
  update_zview( wnd ) ;

  return 0 ;
}

int OnVImgRestoreFulled(GEM_WINDOW* wnd)
{
  VXIMAGE *vimage = (VXIMAGE*) wnd->Extension ;
  float   zratioi ;
  int     xywh[4] ;
  int     xi, yi ;
  int     xe, ye, wxe, whe ;
  int     wxee, whee ;
  int     index ;
  
  for (index = 0; index < 4; index++)
    if ( vimage->zoom_level == zooms[index] ) break ;

  wind_get( wnd->window_handle, WF_PREVXYWH, &xywh[0], &xywh[1], &xywh[2], &xywh[3] ) ;
  wind_calc( WC_WORK, wnd->window_kind, xywh[0], xywh[1], xywh[2], xywh[3], &xe, &ye, &wxe, &whe ) ;

  zratioi = zratio[index] ;
  /* On s'assure que wxe et whe sont multiple du niveau de zoom */
  wxe  = (int)zratioi*(wxe/(int)zratioi);
  whe  = (int)zratioi*(whe/(int)zratioi) ;

  wxee = (int) ((float)wxe/zratioi) ;
  whee = (int) ((float)whe/zratioi) ;
  if (wxee > vimage->inf_img.mfdb.fd_w) wxe = vimage->inf_img.mfdb.fd_w ;
  if (whee > vimage->inf_img.mfdb.fd_h) whe = vimage->inf_img.mfdb.fd_h ;

  wind_calc( WC_BORDER, wnd->window_kind, xe, ye, wxe, whe, &xywh[0], &xywh[1], &xywh[2], &xywh[3] ) ;
  xi = pmfr( vimage->x1, vimage->inf_img.mfdb.fd_w-wxee ) ;
  yi = pmfr( vimage->y1, vimage->inf_img.mfdb.fd_h-whee ) ;
  GWSetHSlider( wnd, xi ) ;
  GWSetVSlider( wnd, yi ) ;
  wxe = pmfr( wxee, vimage->inf_img.mfdb.fd_w ) ;
  whe = pmfr( whee, vimage->inf_img.mfdb.fd_h ) ;
  GWSetHSliderSize( wnd, wxe ) ;
  GWSetVSliderSize( wnd, whe ) ;
  if ( vimage->zoom_level != 1 )
  {
    traite_loupe( wnd, vimage->zoom_level, -1, -1 ) ;
    wnd->must_be_fulled = 1 ;
  }
  else GWOnRestoreFulled( wnd, xywh ) ;

  update_zview( wnd ) ;

  return GW_EVTSTOPROUTING ;
}

int OnVImgIconify(GEM_WINDOW* wnd, int xywh[4])
{
  GWOnIconify( wnd, xywh ) ;
  if ( wnd->window_icon.fd_addr == NULL ) iconify_picture( wnd ) ;

  return 0 ;
}

int OnVImgUnIconify(GEM_WINDOW* wnd, int xywh[4])
{
  GWOnUnIconify( wnd, xywh ) ;
  PostMessage( wnd, WM_TOPPED, NULL ) ;
  update_zview( wnd ) ;

  return GW_EVTSTOPROUTING ;
}

#pragma warn -par
int OnVImgMouseMove(GEM_WINDOW* gwnd, int button, int kstate, int mx, int my)
{
  VXIMAGE *vimage ;
  int     stx, sty ;
  int     ratio ;
  int     index ;
  int     rgb[3] ;
  char    msgxy[50] ;
  char    msgrvb[50] ;
  char    buf[100] ;

  vimage = (VXIMAGE*) gwnd->Extension ;
  if ( gwnd->window_kind & INFO )
  {
    if ( mx >= 0 )
    {
      stx   = vimage->x1 ;
      sty   = vimage->y1 ;
      ratio = vimage->zoom_level ;
      sprintf( msgxy, MsgTxtGetMsg(hMsg, MSG_XY) , stx + mx / ratio, sty + my / ratio ) ;
      index = GetRVBForVIMG( gwnd, mx, my, rgb ) ;
      if ( index < 0 ) sprintf( msgrvb, MsgTxtGetMsg(hMsg, MSG_RVB), rgb[0], rgb[1], rgb[2] ) ;
      else             sprintf( msgrvb, MsgTxtGetMsg(hMsg, MSG_RVBI), rgb[0], rgb[1], rgb[2], index ) ;
      sprintf( buf, "%s %s", msgxy, msgrvb ) ;
      GWSetWindowInfo( gwnd, buf ) ;
    }
    else GWSetWindowInfo( gwnd, "" ) ; /* Sortie de zone */
  }

  return GW_EVTSTOPROUTING ;
}

int OnVImgLButtonDown(GEM_WINDOW* wnd, int mk_state, int x, int y)
{
  if ( wnd->is_icon ) return 0 ;

  traite_action( wnd ) ;

  return GW_EVTSTOPROUTING ;
}

int OnVImgDLButtonDown(GEM_WINDOW* wnd, int mk_state, int x, int y)
{
  if ( wnd->is_icon ) return 0 ;

  return GW_EVTSTOPROUTING ;
}

int OnVImgRButtonDown(GEM_WINDOW* wnd, int mk_state, int x, int y)
{
  traite_iconpopup( wnd ) ;

  return GW_EVTSTOPROUTING ;
}

int OnVImgDRButtonDown(GEM_WINDOW* wnd, int mk_state, int x, int y)
{
  if ( wnd->is_icon ) return 0 ;

  return GW_EVTSTOPROUTING ;
}
#pragma warn +par

int OnVImgProgRange(void* wnd, long val, long max, char* txt)
{
  if ( max )
  {
    short pc = cal_pc( val, max ) ;
    char  buf[100] ;

    if ( ( pc < 100 ) && txt ) sprintf( buf, "%s (%d %%)...", txt, pc ) ;
    else                       buf[0] = 0 ;

    GWSetWindowInfo( wnd, buf ) ;
  }

  return 0 ;
}

static short check_coord(short coord, short maximum)
{
  if ( coord < 0 )            coord = 0 ;
  else if ( coord > maximum ) coord = maximum ;

  return coord ;
}

void ConsistentVClip(void)
{
  int temp ;

  if ( vclip.gwindow )
  {
    VXIMAGE* vimage = (VXIMAGE*) vclip.gwindow->Extension ;

    vclip.x1 = check_coord( vclip.x1, vimage->inf_img.mfdb.fd_w-1 ) ;
    vclip.x2 = check_coord( vclip.x2, vimage->inf_img.mfdb.fd_w-1 ) ;
    vclip.y1 = check_coord( vclip.y1, vimage->inf_img.mfdb.fd_h-1 ) ;
    vclip.y2 = check_coord( vclip.y2, vimage->inf_img.mfdb.fd_h-1 ) ;
  }

  if ( vclip.y1 > vclip.y2 )
  {
    temp     = vclip.y1 ; 
    vclip.y1 = vclip.y2 ;
    vclip.y2 = temp ;
  }
  if ( vclip.x1 > vclip.x2 )
  {
    temp     = vclip.x1 ; 
    vclip.x1 = vclip.x2 ;
    vclip.x2 = temp ;
  }
}